##############################################################################
# 
# Zope Public License (ZPL) Version 0.9.5
# ---------------------------------------
# 
# Copyright (c) Digital Creations.  All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
# 
# 1. Redistributions in source code must retain the above copyright
#    notice, this list of conditions, and the following disclaimer.
# 
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions, and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
# 
# 3. Any use, including use of the Zope software to operate a website,
#    must either comply with the terms described below under
#    "Attribution" or alternatively secure a separate license from
#    Digital Creations.  Digital Creations will not unreasonably
#    deny such a separate license in the event that the request
#    explains in detail a valid reason for withholding attribution.
# 
# 4. All advertising materials and documentation mentioning
#    features derived from or use of this software must display
#    the following acknowledgement:
# 
#      "This product includes software developed by Digital Creations
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
# 
#    In the event that the product being advertised includes an
#    intact Zope distribution (with copyright and license included)
#    then this clause is waived.
# 
# 5. Names associated with Zope or Digital Creations must not be used to
#    endorse or promote products derived from this software without
#    prior written permission from Digital Creations.
# 
# 6. Modified redistributions of any form whatsoever must retain
#    the following acknowledgment:
# 
#      "This product includes software developed by Digital Creations
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
# 
#    Intact (re-)distributions of any official Zope release do not
#    require an external acknowledgement.
# 
# 7. Modifications are encouraged but must be packaged separately as
#    patches to official Zope releases.  Distributions that do not
#    clearly separate the patches from the original work must be clearly
#    labeled as unofficial distributions.  Modifications which do not
#    carry the name Zope may be packaged in any form, as long as they
#    conform to all of the clauses above.
# 
# 
# Disclaimer
# 
#   THIS SOFTWARE IS PROVIDED BY DIGITAL CREATIONS ``AS IS'' AND ANY
#   EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
#   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL DIGITAL CREATIONS OR ITS
#   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
#   USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
#   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
#   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
#   SUCH DAMAGE.
# 
# Attribution
# 
#   Individuals or organizations using this software as a web site must
#   provide attribution by placing the accompanying "button" and a link
#   to the accompanying "credits page" on the website's main entry
#   point.  In cases where this placement of attribution is not
#   feasible, a separate arrangment must be concluded with Digital
#   Creations.  Those using the software for purposes other than web
#   sites must provide a corresponding attribution in locations that
#   include a copyright using a manner best suited to the application
#   environment.  Where attribution is not possible, or is considered
#   to be onerous for some other reason, a request should be made to
#   Digital Creations to waive this requirement in writing.  As stated
#   above, for valid requests, Digital Creations will not unreasonably
#   deny such requests.
# 
# This software consists of contributions made by Digital Creations and
# many individuals on behalf of Digital Creations.  Specific
# attributions are listed in the accompanying credits file.
# 
##############################################################################


from sql import SQLConnection, SQLDataSources, SQLDescribeCol, SQLExecDirect
from sql import SQLFetch, SQLFreeStmt, SQLGetString, SQLNumResultCols
from sql import SQLSetConnectOption, SQLTables, SQLTransact, SQL_AUTOCOMMIT
from sql import SQL_AUTOCOMMIT_OFF, SQL_BIGINT, SQL_BINARY, SQL_BIT, SQL_CHAR
from sql import SQL_COMMIT, SQL_DATE, SQL_DECIMAL, SQL_DOUBLE, SQL_FETCH_FIRST
from sql import SQL_FETCH_NEXT, SQL_FLOAT, SQL_INTEGER, SQL_LONGVARBINARY
from sql import SQL_LONGVARCHAR, SQL_NUMERIC, SQL_REAL, SQL_ROLLBACK
from sql import SQL_SMALLINT, SQL_SUCCESS, SQL_TIME, SQL_TIMESTAMP
from sql import SQL_TINYINT, SQL_VARBINARY, SQL_VARCHAR, SQL_CLOSE

from string import join, split, find, strip
import sys
rc=sys.getrefcount

error=DatabaseError='Solid Database Error'

def manage_ODBCDataSources(*args):
    r=[]
    status, name, desc = SQLDataSources(SQL_FETCH_FIRST)
    while status==SQL_SUCCESS:
	r.append((name, desc))
	status, name, desc = SQLDataSources(SQL_FETCH_NEXT)
    return r

defs={
    SQL_BIGINT: 'i',
    SQL_BINARY: 't',
    SQL_BIT: 'i',
    SQL_CHAR: 't',
    SQL_DATE: 'd',
    SQL_DECIMAL: 'N',
    SQL_DOUBLE: 'n',
    SQL_FLOAT: 'n',
    SQL_INTEGER: 'i',
    SQL_LONGVARBINARY: 't',
    SQL_LONGVARCHAR: 't',
    SQL_NUMERIC: 'N',
    SQL_REAL: 'n',
    SQL_SMALLINT: 'i',
    SQL_TIME: 't',
    SQL_TIMESTAMP: 'd',
    SQL_TINYINT: 'i',
    SQL_VARBINARY: 't',
    SQL_VARCHAR: 't',
    }
tpnames={
    SQL_BIGINT: 'BIGINT',
    SQL_BINARY: 'BINARY',
    SQL_BIT: 'BIT',
    SQL_CHAR: 'CHAR',
    SQL_DATE: 'DATE',
    SQL_DECIMAL: 'DECIMAL',
    SQL_DOUBLE: 'DOUBLE',
    SQL_FLOAT: 'FLOAT',
    SQL_INTEGER: 'INTEGER',
    SQL_LONGVARBINARY: 'LONGVARBINARY',
    SQL_LONGVARCHAR: 'LONGVARCHAR',
    SQL_NUMERIC: 'NUMERIC',
    SQL_REAL: 'REAL',
    SQL_SMALLINT: 'SMALLINT',
    SQL_TIME: 'TIME',
    SQL_TIMESTAMP: 'TIMESTAMP',
    SQL_TINYINT: 'TINYINT',
    SQL_VARBINARY: 'VARBINARY',
    SQL_VARCHAR: 'VARCHAR',
    }

binary_types=SQL_BINARY, SQL_LONGVARBINARY,  SQL_VARBINARY

class DB:

    def __init__(self, connection_string=''):
	self.connection_string=connection_string
	d=split(connection_string)
	if len(d) < 3:  raise 'Invalid Connection String', d
        if d[-1]=='""': d[-1]=''
        if d[-2]=='""': d[-2]=''
        self._connection_info=join(d[0:-2]),d[-2],d[-1]
        self._connect()
        
    def _connect(self):
        self.connection=c=apply(SQLConnection, self._connection_info)
	try: SQLSetConnectOption(c,SQL_AUTOCOMMIT,SQL_AUTOCOMMIT_OFF)
	except: pass # Oh well
        self.stmt=c()

    def _apply(self, f, args):
        # Run a ODBC function with an exception block that
        # Checks for stale connections.
        try: return apply(f, args)
	except error, v:
            try:
                tb=sys.exc_traceback
                state, native, mess = v
                if state[:2]=='08':
                    # Hm, spells like a connection failure.
                    # Let's connect and try again.
                    self._connect()
                    return apply(f, (self.stmt, ) + args[1:] )
                raise error, v, tb
            finally: tb=None

    def table_info(self):
        self._apply(SQLTables,(self.stmt, None, None, None, 'TABLE'))
	r=[]
	for row in split(self.query(), '\n')[2:-1]:
	    q, owner, name, type, remarks = tuple(split(row,'\t'))
	    if owner: owner="(owned by %s) " % owner
	    owner=owner+remarks
	    if owner: owner=', '+owner
	    r.append((name, owner))
	    
	return r

    def tables(self, Qualifier=None, Owner=None, Name=None, Type=None, rdb=1):
	self._apply(SQLTables,(self.stmt, Qualifier, Owner, Name, Type))
	return self.query(rdb=rdb)

    def query(self, src=None, max_rows=99999999, rdb=1):
	stmt=self.stmt
	self._register()
	try:
	    if src is not None:
		r=filter(strip,split(src,'\0'))
		if not r: raise ValueError, 'null sql'
		if len(r) > 1:
		    res=None
		    for s in r:
			self._apply(SQLExecDirect,(stmt,s))
			status, ncol = SQLNumResultCols(stmt)
			if ncol:
			    if res is not None:
				raise ValueError, (
				    'multiple selects are not allowed')
			    res=self.query(None, max_rows, rdb)
		    if res==None:
			if rdb: res="x\n8s\n"
			else: res=()
		    return res

		self._apply(SQLExecDirect,(stmt,src))

	except error, v:
	    state, native, mess = v
	    raise error, "%s (%s)" % (mess, state)

	status, ncol = SQLNumResultCols(stmt)
	if ncol==0:
	    if rdb: return "x\n8s\n"
	    else: return ()
	r=[]
	row=['']*ncol
	names=['']*ncol
	rdbdefs=['']*ncol
	indexes=range(ncol)
	binary_flags=[0]*ncol
	for i in range(ncol):
	    status, name, tp, prec, scale, nullable=SQLDescribeCol(stmt,i+1)
	    names[i]=name
	    if prec < 1: prec=8
	    binary_flags[i]=tp in binary_types
	    if defs.has_key(tp):
                tp=defs[tp]
                if tp=='N' and scale==0: tp='i'
	    else: tp='t'
	    rdbdefs[i]=tp
	    row[i]="%s%s" % (prec,tp)

	if rdb:
	    r.append(join(names,'\t'))
	    r.append(join(row,'\t'))

	status=SQLFetch(stmt)
	while status==SQL_SUCCESS:
	    for i in range(ncol):
		v=SQLGetString(stmt,i+1,binary_flags[i])
		if rdbdefs[i]=='t':
		    if find(v,'\\') >= 0: v=join(split(v,'\\'),'\\\\')
		    if find(v,'\n') >= 0: v=join(split(v,'\n'),'\\n')
		    if find(v,'\t') >= 0: v=join(split(v,'\t'),'\\t')
		row[i]=v
	    if rdb: rd=join(row,'\t')
	    else:
		rd={}
		for i in indexes: rd[names[i]]=row[i]
	    r.append(rd)
	    status=SQLFetch(stmt)

	SQLFreeStmt(stmt, SQL_CLOSE);
	if rdb: r=join(r,'\n')+'\n'
	return r

    def columns(self, table_name):
	stmt=self.stmt
	try: self._apply(SQLExecDirect,
                         (stmt, 'select * from %s where 1=2' % table_name))
	except error, v:
	    state, native, mess = v
	    raise error, "%s (%s)" % (mess, state)

	status, ncol = SQLNumResultCols(stmt)
	r=[]
	standard_type=tpnames.has_key
	for i in range(ncol):
	    status, name, tp, prec, scale, nullable=SQLDescribeCol(stmt,i+1)
	    if standard_type(tp): tp=tpnames[tp]
	    else: tp="Non-standard type %s" % tp
	    r.append({'Name': name, 'Type': tp, 'Precision': prec,
		      'Scale': scale,
		      'Nullable': nullable and 'with Null' or ''})

	SQLFreeStmt(stmt, SQL_CLOSE);
	return r

    class _p_jar:
	# This is place holder for new transaction machinery 2pc
	def __init__(self, db=None): self.db=db
	def begin_commit(self, *args): pass
	def finish_commit(self, *args): pass

    _p_jar=_p_jar(_p_jar())
    _p_oid=_p_changed=_registered=None
    def _register(self):
	if not self._registered:
	    try:
		get_transaction().register(self)
		self._registered=1
	    except: pass

    def __inform_commit__(self, *ignored):
	SQLTransact(self.connection, SQL_COMMIT)
	self._registered=0

    def __inform_abort__(self, *ignored):
	SQLTransact(self.connection, SQL_ROLLBACK)
	self._registered=0

if __name__=='__main__':
    #print manage_ODBCDataSources()
    db=DB('test')
    #print db.table_info()
    if len(sys.argv)==2: print db.query(sys.argv[1])
    else: print db.tables(rdb=0)
    #print db.columns('f')

      
